import os
import tempfile
from typing import NamedTuple

import requests
from dagster import AssetExecutionContext, DagsterInstance, OpExecutionContext
from dagster._annotations import beta
from dagster_cloud_cli.core.errors import raise_http_error
from dagster_cloud_cli.core.headers.auth import DagsterCloudInstanceScope

from dagster_cloud.instance import DagsterCloudAgentInstance


@beta
class DagsterMetric(NamedTuple):
    """Beta: This class gives information about a Metric.

    Args:
        metric_name (str): name of the metric
        metric_value (float): value of the metric
    """

    metric_name: str
    metric_value: float


def get_url_and_token_from_instance(instance: DagsterInstance) -> tuple[str, str]:
    if not isinstance(instance, DagsterCloudAgentInstance):
        raise RuntimeError("This asset only functions in a running Dagster Cloud instance")

    return f"{instance.dagit_url}graphql", instance.dagster_cloud_agent_token


def get_insights_upload_request_params(
    instance: DagsterInstance,
) -> tuple[requests.Session, str, dict[str, str], int, dict[str, str] | None]:
    if not isinstance(instance, DagsterCloudAgentInstance):
        raise RuntimeError("This asset only functions in a running Dagster Cloud instance")

    return (
        instance.requests_managed_retries_session,
        instance.dagster_cloud_gen_insights_url_url,
        instance.dagster_cloud_api_headers(DagsterCloudInstanceScope.DEPLOYMENT),
        instance.dagster_cloud_api_timeout,
        instance.dagster_cloud_api_proxies,
    )


def upload_cost_information(
    context: OpExecutionContext | AssetExecutionContext,
    metric_name: str,
    cost_information: list[tuple[str, float, str]],
):
    import pyarrow as pa
    import pyarrow.parquet as pq

    with tempfile.TemporaryDirectory() as temp_dir:
        opaque_ids = pa.array([opaque_id for opaque_id, _, _ in cost_information])
        costs = pa.array([cost for _, cost, _ in cost_information])
        query_ids = pa.array([query_id for _, _, query_id in cost_information])
        metric_names = pa.array([metric_name for _, _, _ in cost_information])

        cost_pq_file = os.path.join(temp_dir, "cost.parquet")
        pq.write_table(
            pa.Table.from_arrays(
                [opaque_ids, costs, metric_names, query_ids],
                ["opaque_id", "cost", "metric_name", "query_id"],
            ),
            cost_pq_file,
        )

        instance = context.instance
        session, url, headers, timeout, proxies = get_insights_upload_request_params(instance)

        resp = session.get(url, headers=headers, timeout=timeout, proxies=proxies)
        raise_http_error(resp)
        resp_data = resp.json()

        assert "url" in resp_data and "fields" in resp_data, resp_data

        with open(cost_pq_file, "rb") as f:
            session.post(
                resp_data["url"],
                data=resp_data["fields"],
                files={"file": f},
            )


@beta
def put_cost_information(
    context: OpExecutionContext | AssetExecutionContext,
    metric_name: str,
    cost_information: list[tuple[str, float, str]],
    start: float,
    end: float,
) -> None:
    try:
        upload_cost_information(context, metric_name, cost_information)
    except ImportError as e:
        raise Exception(
            "Dagster insights dependencies not installed. Install dagster-cloud[insights] to use this feature."
        ) from e



import json
import os
import queue
import tempfile
import threading
import time
from unittest import mock

import pytest
import requests
import yaml
from click import Command
from dagster._core.test_utils import environ
from dagster_cloud_cli import config_utils
from dagster_cloud_cli.commands.config import SetupAuthMethod, app, app_configure
from dagster_shared.plus.config import DagsterPlusCliConfig
from typer import Context
from typer.testing import CliRunner


@pytest.fixture
def test_config_file():
    with (
        tempfile.TemporaryDirectory() as dagster_cloud_tmpdir,
        tempfile.TemporaryDirectory() as dg_tmpdir,
    ):
        dagster_cloud_config_path = os.path.join(dagster_cloud_tmpdir, "config")
        dg_config_path = os.path.join(dg_tmpdir, "dg.toml")
        with environ(
            {
                "DAGSTER_CLOUD_CLI_CONFIG": dagster_cloud_config_path,
                "DG_CLI_CONFIG": dg_config_path,
            }
        ):
            yield dagster_cloud_config_path


@pytest.fixture
def populated_config(test_config_file):
    with open(test_config_file, "w", encoding="utf8") as f:
        f.write(
            """
organization: acme
default_deployment: staging
user_token: xyz123
        """
        )


@pytest.fixture
def populated_dg_config(test_config_file):
    test_dg_config_file = os.getenv("DG_CLI_CONFIG")
    assert test_dg_config_file is not None
    with open(test_dg_config_file, "w", encoding="utf8") as f:
        f.write(
            """[cli.plus]
organization = "bcme"
default_deployment = "prod"
user_token = "abc123"
        """
        )


def test_read_write_config(test_config_file):
    config = DagsterPlusCliConfig(
        organization="elementl", default_deployment="prod", user_token="abc123"
    )
    config_utils.write_config(config)

    read_config = config_utils.read_config()

    assert read_config.organization == "elementl"
    assert read_config.default_deployment == "prod"
    assert read_config.user_token == "abc123"


def test_overrides(populated_config):
    assert config_utils.get_organization() == "acme"
    assert config_utils.get_deployment() == "staging"
    assert config_utils.get_user_token() == "xyz123"
    assert config_utils.get_url() is None

    with environ(
        {
            "DAGSTER_CLOUD_ORGANIZATION": "elementl",
            "DAGSTER_CLOUD_DEPLOYMENT": "prod",
            "DAGSTER_CLOUD_API_TOKEN": "abc123",
            "DAGSTER_CLOUD_URL": "localhost:3000",
        }
    ):
        assert config_utils.get_organization() == "elementl"
        assert config_utils.get_deployment() == "prod"
        assert config_utils.get_user_token() == "abc123"
        assert config_utils.get_url() == "localhost:3000"

        ctx = Context(Command("test"))
        ctx.params["organization"] = "test"
        ctx.params["deployment"] = "dev"
        ctx.params["api_token"] = "def456"
        ctx.params["url"] = "localhost:4000"

        assert config_utils.get_organization(ctx) == "test"
        assert config_utils.get_deployment(ctx) == "dev"
        assert config_utils.get_user_token(ctx) == "def456"
        assert config_utils.get_url(ctx) == "localhost:4000"


def test_view_command(populated_config):
    runner = CliRunner()
    result = runner.invoke(app, ["view"])

    assert result.exit_code == 0, result.output + " : " + str(result.exception)

    assert yaml.safe_load(result.output) == {
        "organization": "acme",
        "default_deployment": "staging",
        "user_token": "xyz123",
    }


def test_view_command_using_dg_config(populated_dg_config):
    runner = CliRunner()
    result = runner.invoke(app, ["view"])

    assert result.exit_code == 0, result.output + " : " + str(result.exception)

    assert yaml.safe_load(result.output) == {
        "organization": "bcme",
        "default_deployment": "prod",
        "user_token": "abc123",
    }


def test_view_command_using_both_configs(populated_config, populated_dg_config):
    runner = CliRunner()
    result = runner.invoke(app, ["view"])

    # We loudly error if a user has config in both places, to avoid silently performing unexpected behavior
    assert result.exit_code == 1, "Found Dagster Plus config in both" in result.output


# Test setup command, using the CLI auth option
def test_setup_command_cli(test_config_file):
    runner = CliRunner()

    with mock.patch(
        "dagster_cloud_cli.commands.config._settings_method_input",
        mock.Mock(return_value=SetupAuthMethod.CLI),
    ):
        # First, we test the case where a GraphQL call to get the deployments list fails
        # This prompts the user for a text input
        with mock.patch(
            "dagster_cloud_cli.commands.config.gql.fetch_full_deployments",
            mock.Mock(return_value=Exception()),
        ):
            # Ensure initial configure works
            input_mock = mock.Mock(
                side_effect=[
                    "hooli",  # organization
                    "prod",  # deployment
                ]
            )
            password_input_mock = mock.Mock(
                side_effect=[
                    "abc123",  # password
                ],
            )
            with (
                mock.patch("dagster_cloud_cli.commands.config.ui.input", input_mock),
                mock.patch(
                    "dagster_cloud_cli.commands.config.ui.password_input",
                    password_input_mock,
                ),
            ):
                result = runner.invoke(app, ["setup"])

            assert result.exit_code == 0, result.output + " : " + str(result.exception)

            # Verify defaults were as expected
            input_mock.assert_has_calls(
                [mock.call(mock.ANY, default=""), mock.call(mock.ANY, default="")]
            )
            password_input_mock.assert_has_calls([mock.call(mock.ANY, default="")])

            # Verify new configuraiton success
            assert config_utils.get_organization() == "hooli"
            assert config_utils.get_deployment() == "prod"
            assert config_utils.get_user_token() == "abc123"

            # Verify we can reconfigure the CLI, with old defaults
            input_mock = mock.Mock(
                side_effect=[
                    "hooli",  # organization
                    "staging",  # deployment
                ]
            )
            password_input_mock = mock.Mock(
                side_effect=["def456"],  # password
            )
            with (
                mock.patch("dagster_cloud_cli.commands.config.ui.input", input_mock),
                mock.patch(
                    "dagster_cloud_cli.commands.config.ui.password_input",
                    password_input_mock,
                ),
            ):
                result = runner.invoke(app, ["setup"])

            assert result.exit_code == 0, result.output + " : " + str(result.exception)

            # Verify defaults were as expected
            input_mock.assert_has_calls(
                [
                    mock.call(mock.ANY, default="hooli"),
                    mock.call(mock.ANY, default="prod"),
                ]
            )
            password_input_mock.assert_has_calls([mock.call(mock.ANY, default="abc123")])

            # Verify new configuraiton success
            assert config_utils.get_organization() == "hooli"
            assert config_utils.get_deployment() == "staging"
            assert config_utils.get_user_token() == "def456"

        # Next, test when deployment list is returned successfully
        # This prompts a list input, with the default being the existing deployment
        with mock.patch(
            "dagster_cloud_cli.commands.config.gql.fetch_full_deployments",
            mock.Mock(return_value=[{"deploymentName": "prod"}, {"deploymentName": "staging"}]),
        ):
            # Verify we can reconfigure the CLI, with old defaults

            input_mock = mock.Mock(side_effect=["hooli"])  # organization
            password_input_mock = mock.Mock(side_effect=["abc123"])  # user token
            list_input_mock = mock.Mock(side_effect=["prod"])  # deployment

            with (
                mock.patch("dagster_cloud_cli.commands.config.ui.input", input_mock),
                mock.patch(
                    "dagster_cloud_cli.commands.config.ui.password_input",
                    password_input_mock,
                ),
                mock.patch("dagster_cloud_cli.commands.config.ui.list_input", list_input_mock),
            ):
                result = runner.invoke(app, ["setup"])

            assert result.exit_code == 0, result.output + " : " + str(result.exception)

            # Verify defaults were as expected
            input_mock.assert_has_calls([mock.call(mock.ANY, default="hooli")])
            password_input_mock.assert_has_calls([mock.call(mock.ANY, default="def456")])
            list_input_mock.assert_has_calls(
                [mock.call(mock.ANY, choices=["None", "prod", "staging"], default="staging")]
            )

            # Verify new configuraiton success
            assert config_utils.get_organization() == "hooli"
            assert config_utils.get_deployment() == "prod"
            assert config_utils.get_user_token() == "abc123"

        # Sanity check that if the current deployment is not in the list, it is not passed as default
        # Also test backcompat with `dagster-cloud configure`
        with mock.patch(
            "dagster_cloud_cli.commands.config.gql.fetch_full_deployments",
            mock.Mock(
                return_value=[
                    {"deploymentName": "staging"},
                    {"deploymentName": "releasing"},
                ]
            ),
        ):
            input_mock = mock.Mock(side_effect=["elementl"])  # organization
            password_input_mock = mock.Mock(side_effect=["def456"])  # user token
            list_input_mock = mock.Mock(side_effect=["releasing"])  # deployment

            with (
                mock.patch("dagster_cloud_cli.commands.config.ui.input", input_mock),
                mock.patch(
                    "dagster_cloud_cli.commands.config.ui.password_input",
                    password_input_mock,
                ),
                mock.patch("dagster_cloud_cli.commands.config.ui.list_input", list_input_mock),
            ):
                result = runner.invoke(app_configure, [])

            assert result.exit_code == 0, result.output + " : " + str(result.exception)

            # Verify defaults were as expected
            input_mock.assert_has_calls([mock.call(mock.ANY, default="hooli")])
            password_input_mock.assert_has_calls([mock.call(mock.ANY, default="abc123")])
            list_input_mock.assert_has_calls(
                [
                    mock.call(
                        mock.ANY,
                        choices=["None", "staging", "releasing"],
                        default="None",
                    )
                ]
            )

            # Verify new configuraiton success
            assert config_utils.get_organization() == "elementl"
            assert config_utils.get_deployment() == "releasing"
            assert config_utils.get_user_token() == "def456"


# Test setup command, using the web auth option
def test_setup_command_web(test_config_file):
    runner = CliRunner()

    with (
        mock.patch(
            "dagster_shared.utils.find_free_port",
            mock.Mock(return_value=4000),
        ),
        mock.patch(
            "dagster_shared.plus.login_server._generate_nonce",
            mock.Mock(return_value="ABCDEFGH"),
        ),
        mock.patch(
            "dagster_cloud_cli.commands.config.webbrowser.open",
            mock.Mock(return_value=True),
        ),
        mock.patch(
            "dagster_cloud_cli.commands.config._settings_method_input",
            mock.Mock(return_value=SetupAuthMethod.WEB),
        ),
        mock.patch(
            "dagster_cloud_cli.commands.config.gql.fetch_full_deployments",
            mock.Mock(return_value=[{"deploymentName": "prod"}, {"deploymentName": "staging"}]),
        ),
    ):
        assert config_utils.get_organization() is None
        assert config_utils.get_deployment() is None
        assert config_utils.get_user_token() is None

        # Send configuration response to CLI endpoint, HTTP response passed back in queue
        q = queue.Queue()

        def respond_with_callback():
            time.sleep(0.25)
            response = requests.post(
                "http://localhost:4000/callback",
                headers={"Content-Type": "application/json"},
                data=json.dumps({"nonce": "ABCDEFGH", "organization": "hooli", "token": "abc123"}),
            )
            q.put(response)

        th = threading.Thread(target=respond_with_callback)
        th.start()

        # Verify that the CLI is reconfigured by hitting the callback endpoint
        list_input_mock = mock.Mock(side_effect=["prod"])  # deployment

        with mock.patch("dagster_cloud_cli.commands.config.ui.list_input", list_input_mock):
            result = runner.invoke(app, ["setup"])
        th.join()

        assert result.exit_code == 0, result.output + " : " + str(result.exception)

        assert q.get().json().get("ok") is True, "JSON response from callback not as expected"

        # Verify defaults were as expected
        list_input_mock.assert_has_calls(
            [mock.call(mock.ANY, choices=["None", "prod", "staging"], default="None")]
        )

        # Verify new configuration success
        assert config_utils.get_organization() == "hooli"
        assert config_utils.get_deployment() == "prod"
        assert config_utils.get_user_token() == "abc123"


def test_setup_command_interrupt(test_config_file):
    runner = CliRunner()

    # Ensure keyboard interrupt stops execution
    input_mock = mock.Mock(
        side_effect=[
            "hooli",  # organization
        ]
    )
    with (
        mock.patch("dagster_cloud_cli.commands.config.ui.input", input_mock),
        mock.patch(
            "dagster_cloud_cli.commands.config.ui.password_input",
            side_effect=KeyboardInterrupt,
        ),
    ):
        result = runner.invoke(app, ["setup"])

    assert result.exit_code == 1, result.output + " : " + str(result.exception)

    # Verify no new configuraiton
    assert config_utils.get_organization() is None
    assert config_utils.get_deployment() is None
    assert config_utils.get_user_token() is None


def test_set_deployment_command(populated_config):
    with mock.patch(
        "dagster_cloud_cli.commands.config.available_deployment_names"
    ) as available_deployment_names:
        assert config_utils.get_deployment() == "staging"
        available_deployment_names.return_value = ["prod", "staging"]

        # Verify we can switch deployment to prod
        runner = CliRunner()
        result = runner.invoke(app, ["set-deployment", "prod"])

        assert result.exit_code == 0, result.output + " : " + str(result.exception)
        assert config_utils.get_deployment() == "prod"

        # Verify we can't switch deployment to fake
        runner = CliRunner()
        result = runner.invoke(app, ["set-deployment", "fake"])

        assert result.exit_code == 1, result.output + " : " + str(result.exception)
        assert config_utils.get_deployment() == "prod"



from dagster_cloud.workspace.ecs.client import Client
from dagster_cloud.workspace.ecs.service import Service


def test_arn():
    short_arn = "arn:aws:ecs:us-east-1:1234567890:service/service-name"
    long_arn = "arn:aws:ecs:us-east-1:1234567890:service/cluster-name/service-name"

    client = Client(
        cluster_name="cluster-name",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
    )

    service = Service(client=client, arn=short_arn)
    assert service.arn == long_arn

    service = Service(client=client, arn=long_arn)
    assert service.arn == long_arn



import copy
import socket
import uuid
from collections.abc import Sequence
from contextlib import ExitStack
from functools import lru_cache
from typing import TYPE_CHECKING, Any

import yaml
from dagster import (
    Array,
    Field,
    Map,
    String,
    _check as check,
)
from dagster._config import process_config
from dagster._core.errors import DagsterInvalidConfigError, DagsterInvariantViolationError
from dagster._core.instance import DagsterInstance
from dagster._core.instance.config import config_field_for_configurable_class
from dagster._core.instance.ref import InstanceRef, configurable_class_data
from dagster._core.launcher import DefaultRunLauncher, RunLauncher
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.storage.defs_state.base import DefsStateStorage
from dagster._serdes import ConfigurableClassData
from dagster_cloud_cli.core.graphql_client import (
    create_agent_graphql_client,
    create_agent_http_client,
    create_graphql_requests_session,
    get_agent_headers,
)
from dagster_cloud_cli.core.headers.auth import DagsterCloudInstanceScope
from urllib3 import Retry

from dagster_cloud.agent import AgentQueuesConfig
from dagster_cloud.auth.constants import decode_agent_token
from dagster_cloud.storage.client import dagster_cloud_api_config
from dagster_cloud.util import get_env_names_from_config, is_isolated_run

if TYPE_CHECKING:
    from requests import Session

    from dagster_cloud.workspace.user_code_launcher.user_code_launcher import (
        DagsterCloudUserCodeLauncher,
    )


class DagsterCloudInstance(DagsterInstance):
    @property
    def telemetry_enabled(self) -> bool:
        return False

    @property
    def run_retries_max_retries(self) -> int:
        raise NotImplementedError(
            "run_retries.max_retries is a deployment setting and can only be accessed by a DeploymentScopedHostInstance"
        )

    @property
    def run_retries_retry_on_asset_or_op_failure(self) -> bool:
        raise NotImplementedError(
            "run_retries.retry_on_asset_or_op_failure is a deployment setting and can only be accessed by a DeploymentScopedHostInstance"
        )

    @property
    def defs_state_storage(self) -> DefsStateStorage | None:
        # only DeploymentScopedHostInstance / DagsterCloudAgentInstance have a defs state storage
        return None


class DagsterCloudAgentInstance(DagsterCloudInstance):
    def __init__(
        self,
        *args,
        dagster_cloud_api,
        user_code_launcher=None,
        agent_replicas=None,
        isolated_agents=None,
        agent_queues=None,
        allowed_full_deployment_locations=None,
        allowed_branch_deployment_locations=None,
        agent_metrics=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._unprocessed_dagster_cloud_api_config = dagster_cloud_api
        self._dagster_cloud_api_config = check.not_none(
            self._get_processed_config(
                "dagster_cloud_api",
                dagster_cloud_api,
                dagster_cloud_api_config(),
            )
        )

        check.invariant(
            not (
                self._dagster_cloud_api_config.get("deployment")
                and self._dagster_cloud_api_config.get("deployments")
            ),
            "Cannot set both deployment and deployments in `dagster_cloud_api`",
        )

        self._user_code_launcher_data = (
            configurable_class_data(user_code_launcher) if user_code_launcher else None
        )

        if not self._user_code_launcher_data:
            # This is a user facing error. We should have more actionable advice and link to docs here.
            raise DagsterInvariantViolationError(
                "User code launcher is not configured for DagsterCloudAgentInstance. Configure a"
                " user code launcher under the user_code_launcher: key in your dagster.yaml file."
            )

        self._exit_stack = ExitStack()

        self._user_code_launcher = None
        self._graphql_requests_session: Session | None = None
        self._rest_requests_session: Session | None = None
        self._graphql_client = None
        self._http_client = None

        assert self.dagster_cloud_url

        # Handle backcompat between isolated_agents and agent_replicas
        if isolated_agents and agent_replicas:
            raise Exception(
                "Cannot provide both isolated_agents and agent_replicas configuration. Please only"
                " provide one of these."
            )
        if agent_replicas:
            self._isolated_agents = self._get_processed_config(
                "agent_replicas", agent_replicas, self._isolated_agents_config_schema()
            )
        elif isolated_agents:
            self._isolated_agents = self._get_processed_config(
                "isolated_agents", isolated_agents, self._isolated_agents_config_schema()
            )
        else:
            self._isolated_agents = None

        processed_agent_queues_config = self._get_processed_config(
            "agent_queues", agent_queues, self._agent_queues_config_schema()
        )
        self.agent_queues_config = AgentQueuesConfig(**processed_agent_queues_config)  # pyright: ignore[reportCallIssue]

        self._allowed_full_deployment_locations: dict[str, list[str]] | None = (
            allowed_full_deployment_locations
        )
        self._allowed_branch_deployment_locations: list[str] | None = (
            allowed_branch_deployment_locations
        )

        self._instance_uuid = str(uuid.uuid4())

    @property
    def defs_state_storage(self) -> DefsStateStorage | None:
        # temporary hack to avoid cases where the default BlobStorageStateStorage is used
        from dagster_cloud.storage.defs_state.storage import GraphQLDefsStateStorage

        return (
            self._defs_state_storage
            if isinstance(self._defs_state_storage, GraphQLDefsStateStorage)
            else None
        )

    def _get_processed_config(
        self, name: str, config: dict[str, Any] | None, config_type: dict[str, Any]
    ):
        config_dict = check.opt_dict_param(config, "config", key_type=str)
        processed_config = process_config(config_type, config_dict)
        if not processed_config.success:
            raise DagsterInvalidConfigError(
                f"Errors whilst loading {name} config",
                processed_config.errors,
                config_dict,
            )
        return processed_config.value

    def _dagster_cloud_api_config_for_deployment(self, deployment_name: str | None):
        new_api_config = dict(copy.deepcopy(self._dagster_cloud_api_config))
        if deployment_name:
            new_api_config["deployment"] = deployment_name
            if self.includes_branch_deployments:
                del new_api_config["branch_deployments"]
            new_api_config.pop("deployments", None)
            new_api_config.pop("all_serverless_deployments", None)
        else:
            new_api_config.pop("deployment", None)
            new_api_config.pop("deployments", None)

        return new_api_config

    def ref_for_deployment(self, deployment_name: str) -> InstanceRef:
        my_ref = self.get_ref()
        my_custom_instance_class_data = my_ref.custom_instance_class_data
        new_class_data = _cached_inject_deployment(my_custom_instance_class_data, deployment_name)

        return my_ref._replace(custom_instance_class_data=new_class_data)

    def organization_scoped_graphql_client(self):
        return create_agent_graphql_client(
            self.client_managed_retries_requests_session,
            self.dagster_cloud_graphql_url,
            self._dagster_cloud_api_config_for_deployment(None),
            scope=DagsterCloudInstanceScope.ORGANIZATION,
        )

    def graphql_client_for_deployment(self, deployment_name: str | None):
        return create_agent_graphql_client(
            self.client_managed_retries_requests_session,
            self.dagster_cloud_graphql_url,
            self._dagster_cloud_api_config_for_deployment(deployment_name),
            scope=DagsterCloudInstanceScope.DEPLOYMENT,
        )

    def headers_for_deployment(self, deployment_name: str):
        return get_agent_headers(
            self._dagster_cloud_api_config_for_deployment(deployment_name),
            DagsterCloudInstanceScope.DEPLOYMENT,
        )

    def create_graphql_client(
        self, scope: DagsterCloudInstanceScope = DagsterCloudInstanceScope.DEPLOYMENT
    ):
        return create_agent_graphql_client(
            self.client_managed_retries_requests_session,
            self.dagster_cloud_graphql_url,
            self._dagster_cloud_api_config,  # pyright: ignore[reportArgumentType]
            scope=scope,
        )

    def _translate_socket_param(self, socket_option: str | int):
        if isinstance(socket_option, str):
            check.invariant(
                hasattr(socket, socket_option),
                f"socket module does not have an {socket_option} attribute",
            )
            return getattr(socket, socket_option)
        else:
            return socket_option

    def _socket_options(self):
        if self._dagster_cloud_api_config.get("socket_options") is None:
            return None

        translated_socket_options = []
        for socket_option in self._dagster_cloud_api_config["socket_options"]:
            check.invariant(
                len(socket_option) == 3, "Each socket option must be a list of three values"
            )
            socket_param_1, socket_param_2, socket_val = socket_option
            translated_socket_options.append(
                (
                    self._translate_socket_param(socket_param_1),
                    self._translate_socket_param(socket_param_2),
                    socket_val,
                )
            )

        return translated_socket_options

    @property
    def client_managed_retries_requests_session(self):
        """A shared requests Session to use between GraphQL clients.

        Retries handled in GraphQL client layer.
        """
        if self._graphql_requests_session is None:
            self._graphql_requests_session = self._exit_stack.enter_context(
                create_graphql_requests_session(
                    adapter_kwargs=dict(socket_options=self._socket_options())
                )
            )

        return self._graphql_requests_session

    @property
    def requests_managed_retries_session(self):
        """A requests session to use for non-GraphQL Rest API requests.

        Retries handled by requests.
        """
        if self._rest_requests_session is None:
            self._rest_requests_session = self._exit_stack.enter_context(
                create_graphql_requests_session(
                    adapter_kwargs=dict(
                        max_retries=Retry(
                            total=self.dagster_cloud_api_retries,
                            backoff_factor=self._dagster_cloud_api_config["backoff_factor"],
                        ),
                        socket_options=self._socket_options(),
                    )
                )
            )
        return self._rest_requests_session

    @property
    def graphql_client(self):
        if self._graphql_client is None:
            self._graphql_client = self.create_graphql_client(
                scope=DagsterCloudInstanceScope.DEPLOYMENT
            )

        return self._graphql_client

    @property
    def http_client(self):
        if self._http_client is None:
            self._http_client = create_agent_http_client(
                self.client_managed_retries_requests_session,
                self._dagster_cloud_api_config,  # pyright: ignore[reportArgumentType]
                scope=DagsterCloudInstanceScope.DEPLOYMENT,
            )

        return self._http_client

    @property
    def dagster_cloud_url(self):
        if "url" in self._dagster_cloud_api_config:
            return self._dagster_cloud_api_config["url"]

        organization, region = decode_agent_token(self.dagster_cloud_agent_token)
        if not organization:
            raise DagsterInvariantViolationError(
                "Could not derive Dagster Cloud URL from agent token. Create a new agent token or"
                " set the `url` field under `dagster_cloud_api` in your `dagster.yaml`."
            )

        return (
            f"https://{organization}.agent.{region}.dagster.cloud"
            if region
            else f"https://{organization}.agent.dagster.cloud"
        )

    @property
    def organization_name(self) -> str | None:
        organization, _ = decode_agent_token(self.dagster_cloud_agent_token)
        return organization

    @property
    def deployment_name(self) -> str | None:
        deployment_names = self.deployment_names
        check.invariant(
            len(deployment_names) <= 1,
            "Cannot call instance.deployment_name if multiple deployments are set",
        )
        if not deployment_names:
            return None
        return deployment_names[0]

    @property
    def deployment_names(self) -> list[str]:
        if self._dagster_cloud_api_config.get("deployment"):
            return [self._dagster_cloud_api_config["deployment"]]

        return self._dagster_cloud_api_config.get("deployments", [])

    @property
    def include_all_serverless_deployments(self) -> bool:
        return self._dagster_cloud_api_config.get("all_serverless_deployments") or False

    @property
    def dagit_url(self):
        organization, region = decode_agent_token(self.dagster_cloud_agent_token)
        if not organization:
            raise Exception(
                "Could not derive Dagster Cloud URL from agent token to generate a Dagit URL."
                " Generate a new agent token in the Dagit UI."
            )

        deployment = self._dagster_cloud_api_config.get("deployment")
        base_url = (
            f"https://{organization}.{region}.dagster.cloud/"
            if region
            else f"https://{organization}.dagster.cloud/"
        )
        return base_url + (f"{deployment}/" if deployment else "")

    @property
    def dagster_cloud_graphql_url(self):
        return f"{self.dagster_cloud_url}/graphql"

    @property
    def dagster_cloud_store_events_url(self):
        return f"{self.dagster_cloud_url}/store_events"

    @property
    def dagster_cloud_upload_logs_url(self):
        return f"{self.dagster_cloud_url}/upload_logs"

    @property
    def dagster_cloud_gen_logs_url_url(self):
        return f"{self.dagster_cloud_url}/gen_logs_url"

    @property
    def dagster_cloud_gen_insights_url_url(self) -> str:
        return f"{self.dagster_cloud_url}/gen_insights_url"

    @property
    def dagster_cloud_upload_job_snap_url(self):
        return f"{self.dagster_cloud_url}/upload_job_snapshot"

    @property
    def dagster_cloud_upload_workspace_entry_url(self):
        return f"{self.dagster_cloud_url}/upload_workspace_entry"

    @property
    def dagster_cloud_upload_api_response_url(self):
        return f"{self.dagster_cloud_url}/upload_api_response"

    @property
    def dagster_cloud_check_snapshot_url(self):
        return f"{self.dagster_cloud_url}/check_snapshot"

    @property
    def dagster_cloud_confirm_upload_url(self):
        return f"{self.dagster_cloud_url}/confirm_upload"

    @property
    def dagster_cloud_code_location_update_result_url(self):
        return f"{self.dagster_cloud_url}/code_location_update_result"

    def dagster_cloud_api_headers(self, scope: DagsterCloudInstanceScope):
        return get_agent_headers(self._dagster_cloud_api_config, scope=scope)  # pyright: ignore[reportArgumentType]

    @property
    def dagster_cloud_agent_token(self) -> str:
        check.invariant(
            self._dagster_cloud_api_config.get("agent_token") is not None,
            "No agent token found in dagster_cloud_api configuration. An agent token is required"
            " for Dagster Cloud authentication.",
        )
        return self._dagster_cloud_api_config["agent_token"]

    @property
    def dagster_cloud_api_retries(self) -> int:
        return self._dagster_cloud_api_config["retries"]

    @property
    def dagster_cloud_api_timeout(self) -> int:
        return self._dagster_cloud_api_config["timeout"]

    @property
    def dagster_cloud_api_proxies(self) -> dict[str, str] | None:
        # Requests library modifies the proxies key so create a copy
        return (
            self._dagster_cloud_api_config.get("proxies").copy()  # pyright: ignore[reportOptionalMemberAccess]
            if self._dagster_cloud_api_config.get("proxies")
            else {}
        )

    @property
    def dagster_cloud_api_agent_label(self) -> str | None:
        return self._dagster_cloud_api_config.get("agent_label")

    @property
    def includes_branch_deployments(self) -> bool:
        return self._dagster_cloud_api_config.get("branch_deployments", False)

    @property
    def instance_uuid(self) -> str:
        return self._instance_uuid

    @property
    def allowed_full_deployment_locations(self) -> dict[str, list[str]] | None:
        return self._allowed_full_deployment_locations

    @property
    def allowed_branch_deployment_locations(self) -> list[str] | None:
        return self._allowed_branch_deployment_locations

    def is_location_allowed(
        self, deployment_name: str, location_name: str, is_branch_deployment: bool
    ) -> bool:
        """Check if a location should be allowed based on the allowed locations configuration."""
        if is_branch_deployment:
            # For branch deployments, check the branch deployment locations list
            if self._allowed_branch_deployment_locations is not None:
                return location_name in self._allowed_branch_deployment_locations
        else:
            # For full deployments, check the deployment-specific locations map
            if self._allowed_full_deployment_locations is not None:
                allowed_locations = self._allowed_full_deployment_locations.get(deployment_name)
                if allowed_locations is not None:
                    return location_name in allowed_locations

        # If no restrictions are configured, allow all locations
        return True

    @property
    def agent_display_name(self) -> str:
        if self.dagster_cloud_api_agent_label:
            return f"Agent {self.instance_uuid[:8]} ({self.dagster_cloud_api_agent_label})"
        else:
            return f"Agent {self.instance_uuid[:8]}"

    @property
    def dagster_cloud_api_env_vars(self) -> list[str]:
        return get_env_names_from_config(
            dagster_cloud_api_config(), self._unprocessed_dagster_cloud_api_config
        )

    @property
    def user_code_launcher(self) -> "DagsterCloudUserCodeLauncher":
        # Lazily load in case the user code launcher requires dependencies (like dagster-k8s)
        # that we don't neccesarily need to load in every context that loads a
        # DagsterCloudAgentInstance (for example, a step worker)
        from dagster_cloud.workspace.user_code_launcher.user_code_launcher import (
            DagsterCloudUserCodeLauncher,
        )

        if not self._user_code_launcher:
            self._user_code_launcher = self._exit_stack.enter_context(
                self._user_code_launcher_data.rehydrate(as_type=DagsterCloudUserCodeLauncher)  # type: ignore  # (possible none)
            )
            self._user_code_launcher.register_instance(self)
        return self._user_code_launcher

    @property
    def run_launcher(self) -> RunLauncher:
        """The agent has two run launchers, isolated and non-isolated. Use get_run_launcher_for_run to
        get the appropriate one. This method will always return the isolated run launcher, which
        is required for some OSS peices like the k8s executor.
        """
        return self.user_code_launcher.run_launcher()

    def get_run_launcher_for_run(self, run: DagsterRun) -> RunLauncher:
        # If the run is isolated, use the isolated run launcher, which is specific to whatever agent
        # type we're using- ECS, K8s, etc.
        #
        # If the run is not isolated, use the DefaultRunLauncher to send it to the gRPC server.
        if is_isolated_run(run):
            return self.user_code_launcher.run_launcher()
        else:
            launcher = DefaultRunLauncher()
            launcher.register_instance(self)
            return launcher

    @staticmethod
    def get():
        instance = DagsterInstance.get()
        if not isinstance(instance, DagsterCloudAgentInstance):
            raise DagsterInvariantViolationError(
                """
DagsterInstance.get() did not return a DagsterCloudAgentInstance. Make sure that your"
`dagster.yaml` file is correctly configured to include the following:
instance_class:
  module: dagster_cloud.instance
  class: DagsterCloudAgentInstance
"""
            )
        return instance

    @classmethod
    def config_schema(cls):
        return {
            "dagster_cloud_api": Field(dagster_cloud_api_config(), is_required=True),
            "user_code_launcher": config_field_for_configurable_class(),
            "isolated_agents": Field(cls._isolated_agents_config_schema(), is_required=False),
            "agent_replicas": Field(
                cls._isolated_agents_config_schema(), is_required=False
            ),  # deprecated in favor of isolated_agents
            "agent_queues": Field(cls._agent_queues_config_schema(), is_required=False),
            "allowed_full_deployment_locations": Field(
                Map(String, Array(String)),
                is_required=False,
                description="Mapping of full deployment names to allowed location names",
            ),
            "allowed_branch_deployment_locations": Field(
                Array(String),
                is_required=False,
                description="List of allowed location names for branch deployments",
            ),
        }

    @classmethod
    def _code_server_metrics_config_schema(cls):
        return {"enabled": Field(bool, is_required=False, default_value=False)}

    @classmethod
    def _isolated_agents_config_schema(cls):
        return {"enabled": Field(bool, is_required=False, default_value=False)}

    @classmethod
    def _agent_queues_config_schema(cls):
        return {
            "include_default_queue": Field(bool, default_value=True),
            "additional_queues": Field(Array(String), is_required=False),
        }

    def get_required_daemon_types(self) -> Sequence[str]:
        return []

    @staticmethod
    def config_defaults(base_dir):
        defaults = InstanceRef.config_defaults(base_dir)

        empty_yaml = yaml.dump({})

        defaults["run_storage"] = ConfigurableClassData(  # pyright: ignore[reportIndexIssue]
            "dagster_cloud.storage.runs",
            "GraphQLRunStorage",
            empty_yaml,
        )
        defaults["event_log_storage"] = ConfigurableClassData(  # pyright: ignore[reportIndexIssue]
            "dagster_cloud.storage.event_logs",
            "GraphQLEventLogStorage",
            empty_yaml,
        )
        defaults["schedule_storage"] = ConfigurableClassData(  # pyright: ignore[reportIndexIssue]
            "dagster_cloud.storage.schedules",
            "GraphQLScheduleStorage",
            empty_yaml,
        )
        defaults["storage"] = ConfigurableClassData(  # pyright: ignore[reportIndexIssue]
            module_name="dagster.core.storage.legacy_storage",
            class_name="CompositeStorage",
            config_yaml=yaml.dump(
                {
                    "run_storage": {
                        "module_name": "dagster_cloud.storage.runs",
                        "class_name": "GraphQLRunStorage",
                        "config_yaml": empty_yaml,
                    },
                    "event_log_storage": {
                        "module_name": "dagster_cloud.storage.event_logs",
                        "class_name": "GraphQLEventLogStorage",
                        "config_yaml": empty_yaml,
                    },
                    "schedule_storage": {
                        "module_name": "dagster_cloud.storage.schedules",
                        "class_name": "GraphQLScheduleStorage",
                        "config_yaml": empty_yaml,
                    },
                },
                default_flow_style=False,
            ),
        )

        defaults["compute_logs"] = ConfigurableClassData(  # pyright: ignore[reportIndexIssue]
            "dagster_cloud.storage.compute_logs", "CloudComputeLogManager", empty_yaml
        )

        defaults["secrets"] = ConfigurableClassData(  # pyright: ignore[reportIndexIssue]
            "dagster_cloud.secrets", "DagsterCloudSecretsLoader", empty_yaml
        )

        defaults["defs_state_storage"] = ConfigurableClassData(  # pyright: ignore[reportIndexIssue]
            "dagster_cloud.storage.defs_state", "GraphQLDefsStateStorage", empty_yaml
        )

        return defaults

    def dispose(self) -> None:
        super().dispose()
        self._exit_stack.close()

    @property
    def should_start_background_run_thread(self) -> bool:
        # If using isolated agents (that is, agents cannot see each other's grpc
        # servers), TERMINATE_RUN requests can't be served by the Dagster Cloud API,
        # which will send requests to any agent. Only the agent with the run can
        # terminate it - so we need to start a background run thread to monitor for
        # runs moved into a canceling state.
        return self.is_using_isolated_agents

    @property
    def is_using_isolated_agents(self) -> bool:
        return self._isolated_agents is not None and self._isolated_agents.get("enabled", False)

    @property
    def dagster_cloud_run_worker_monitoring_interval_seconds(self) -> int:
        # potentially overridden interval in the serverless user code launcher
        return 30


@lru_cache(maxsize=100)  # Scales on order of active branch deployments
def _cached_inject_deployment(
    custom_instance_class_data: ConfigurableClassData,
    deployment_name: str,
) -> ConfigurableClassData:
    # incurs costly yaml parse
    config_dict = custom_instance_class_data.config_dict

    config_dict["dagster_cloud_api"]["deployment"] = deployment_name
    if config_dict["dagster_cloud_api"].get("branch_deployments"):
        del config_dict["dagster_cloud_api"]["branch_deployments"]
    if config_dict["dagster_cloud_api"].get("deployments"):
        del config_dict["dagster_cloud_api"]["deployments"]

    return ConfigurableClassData(
        "dagster_cloud.instance",
        "DagsterCloudAgentInstance",
        # incurs costly yaml dump
        yaml.dump(config_dict),
    )



from dagster import AssetKey, SourceAsset, asset


@asset
def little_richard():
    pass


def make_list_of_assets():
    # these assets are stashed inside a function so that they need to be discovered through
    # a list
    @asset
    def james_brown():
        pass

    @asset(metadata={"foo": "bar"})
    def fats_domino():
        pass

    return [james_brown, fats_domino]


def make_list_of_source_assets():
    # these source assets are stashed inside a function so that they need to be discovered through
    # a list
    buddy_holly = SourceAsset(key=AssetKey("buddy_holly"))
    jerry_lee_lewis = SourceAsset(key=AssetKey("jerry_lee_lewis"))

    return [buddy_holly, jerry_lee_lewis]


list_of_assets_and_source_assets = [*make_list_of_assets(), *make_list_of_source_assets()]



from dagster_cloud.api.dagster_cloud_api import DagsterCloudApiThreadTelemetry


def test_thread_telemetry():
    thread_telemetry = DagsterCloudApiThreadTelemetry(
        submitted_to_executor_timestamp=0.0,
        thread_start_run_timestamp=0.5,
        thread_end_handle_api_request_timestamp=3.0,
    )

    assert thread_telemetry.time_to_thread_initialization_seconds == 0.5
    assert thread_telemetry.time_to_handle_api_request_seconds == 2.5



import os
from contextlib import nullcontext
from datetime import datetime

import pytest
from dagster import AssetExecutionContext, materialize
from dagster_cloud.dagster_insights import dbt_with_bigquery_insights, dbt_with_snowflake_insights
from dagster_cloud.dagster_insights.bigquery.bigquery_utils import (
    BIGQUERY_METADATA_BYTES_BILLED,
    BIGQUERY_METADATA_SLOTS_MS,
)
from dagster_cloud.dagster_insights.snowflake.dagster_snowflake_insights import (
    get_cost_data_for_hour,
)
from dagster_cloud.dagster_insights.snowflake.snowflake_utils import OPAQUE_ID_METADATA_KEY_PREFIX
from dagster_dbt import DbtCliResource, dbt_assets

from dagster_cloud_tests.dagster_insights_tests.conftest import (
    JAFFLE_SHOP_ASSET_KEYS,
    bigquery_client,
)


@pytest.mark.parametrize(
    "use_iterator_chain",
    [True, False],
)
def test_snowflake_asset_metadata_added(
    snowflake_manifest_path, snowflake_jaffle_dir, use_iterator_chain: bool
):
    @dbt_assets(manifest=snowflake_manifest_path)
    def jaffle_shop_dbt_assets(
        context: AssetExecutionContext,
        dbt: DbtCliResource,
    ):
        if use_iterator_chain:
            yield from (
                dbt.cli(["build", "--profile", "unittest"], context=context)
                .stream()
                .with_insights()
            )
        else:
            dbt_cli_invocation = dbt.cli(["build", "--profile", "unittest"], context=context)
            yield from dbt_with_snowflake_insights(context, dbt_cli_invocation)

    result = materialize(
        [jaffle_shop_dbt_assets],
        resources={
            "dbt": DbtCliResource(project_dir=os.fspath(snowflake_jaffle_dir)),
        },
    )

    asset_keys_with_metadata = set()
    for event in result.all_events:
        if event.event_type != "ASSET_OBSERVATION":
            continue
        for key in event.asset_observation_data.asset_observation.metadata.keys():
            if key.startswith(OPAQUE_ID_METADATA_KEY_PREFIX):
                asset_keys_with_metadata.add(
                    event.asset_observation_data.asset_observation.asset_key.to_user_string()
                )

    assert asset_keys_with_metadata == {
        "customers",
        "orders",
        "raw_customers",
        "raw_orders",
        "raw_payments",
        "stg_customers",
        "stg_orders",
        "stg_payments",
    }


def test_retrieving_cost_data():
    class MockResult:
        def __init__(self, result):
            self._result = result

        def fetchall(self):
            return self._result

    class MockSnowflakeResource:
        def get_connection(self):
            return nullcontext(self)

        def cursor(self):
            return nullcontext(self)

        def execute(self, query: str):
            assert "HAVING ARRAY_SIZE(opaque_ids)" in query
            return MockResult(
                [
                    [
                        '["model.jaffle_shop.customers:ed530106-5713-4dfe-8b34-e206cb9eed3e"]',
                        1,
                        "queryid1",
                    ],
                    [
                        '["model.jaffle_shop.orders:ed530106-5713-4dfe-8b34-e206cb9eed3e"]',
                        2,
                        "queryid2",
                    ],
                    [
                        '["model.jaffle_shop.stg_customers:ed530106-5713-4dfe-8b34-e206cb9eed3e"]',
                        3,
                        "queryid3",
                    ],
                ]
            )

    snowflake = MockSnowflakeResource()
    start_hour = datetime(2023, 1, 1, 1)
    end_hour = datetime(2023, 1, 1, 2)

    results = get_cost_data_for_hour(snowflake, start_hour, end_hour)  # pyright: ignore[reportArgumentType]
    assert set(results) == {
        ("model.jaffle_shop.customers:ed530106-5713-4dfe-8b34-e206cb9eed3e", 1.0, "queryid1"),
        ("model.jaffle_shop.orders:ed530106-5713-4dfe-8b34-e206cb9eed3e", 2.0, "queryid2"),
        ("model.jaffle_shop.stg_customers:ed530106-5713-4dfe-8b34-e206cb9eed3e", 3.0, "queryid3"),
    }


@pytest.mark.parametrize(
    "use_iterator_chain",
    [True, False],
)
def test_bigquery_asset_metadata_added(
    bigquery_manifest_path,
    bigquery_jaffle_dir,
    default_bigquery_client,
    use_iterator_chain: bool,
):
    @dbt_assets(manifest=bigquery_manifest_path)
    def jaffle_shop_dbt_assets(
        context: AssetExecutionContext,
        dbt: DbtCliResource,
    ):
        if use_iterator_chain:
            yield from (
                dbt.cli(["build", "--profile", "unittest"], context=context)
                .stream()
                .with_insights()
            )
        else:
            dbt_cli_invocation = dbt.cli(["build", "--profile", "unittest"], context=context)
            yield from dbt_with_bigquery_insights(context, dbt_cli_invocation)

    result = materialize(
        [jaffle_shop_dbt_assets],
        resources={
            "dbt": DbtCliResource(project_dir=os.fspath(bigquery_jaffle_dir)),
        },
    )

    asset_keys_with_metadata = set()
    for event in result.get_asset_observation_events():
        for key in event.asset_observation_data.asset_observation.metadata.keys():
            if key == BIGQUERY_METADATA_BYTES_BILLED or key == BIGQUERY_METADATA_SLOTS_MS:
                asset_keys_with_metadata.add(
                    event.asset_observation_data.asset_observation.asset_key.to_user_string()
                )

    assert asset_keys_with_metadata == JAFFLE_SHOP_ASSET_KEYS


@pytest.mark.parametrize(
    "bq_client_args",
    [
        ("US", "foo", None, None, "`foo`.`region-us`.INFORMATION_SCHEMA.JOBS"),
        (None, None, "US", "foo", "`foo`.`region-us`.INFORMATION_SCHEMA.JOBS"),
        ("EU", "bar", "US", "foo", "`bar`.`region-eu`.INFORMATION_SCHEMA.JOBS"),
        (None, None, None, None, None),
        (None, "foo", None, None, None),
        ("US", None, None, None, None),
    ],
)
def test_bigquery_client(bigquery_manifest_path, bigquery_jaffle_dir, bq_client_args):
    @dbt_assets(manifest=bigquery_manifest_path)
    def jaffle_shop_dbt_assets(
        context: AssetExecutionContext,
        dbt: DbtCliResource,
    ):
        dbt_cli_invocation = dbt.cli(["build", "--profile", "unittest"], context=context)
        yield from dbt_with_bigquery_insights(context, dbt_cli_invocation)

    client_location, client_project, dataset_location, dataset_project, expected_cost_table = (
        bq_client_args
    )
    with bigquery_client(
        client_location=client_location,
        client_project=client_project,
        dataset_location=dataset_location,
        dataset_project=dataset_project,
    ) as client:
        materialize(
            [jaffle_shop_dbt_assets],
            resources={
                "dbt": DbtCliResource(project_dir=os.fspath(bigquery_jaffle_dir)),
            },
        )

    if expected_cost_table is None:
        assert client.query.call_count == 0
    else:
        assert client.query.call_count == 1
        query_arg = client.query.call_args[0][0]
        assert f"FROM {expected_cost_table}" in query_arg


def test_bigquery_with_check_failures_cost_metadata_added(
    bigquery_with_check_failure_manifest_path,
    bigquery_with_check_failure_jaffle_dir,
    default_bigquery_client,
):
    @dbt_assets(manifest=bigquery_with_check_failure_manifest_path)
    def jaffle_shop_dbt_assets(
        context: AssetExecutionContext,
        dbt: DbtCliResource,
    ):
        dbt_cli_invocation = dbt.cli(["build", "--profile", "unittest"], context=context)
        yield from dbt_with_bigquery_insights(context, dbt_cli_invocation)

    result = materialize(
        [jaffle_shop_dbt_assets],
        resources={
            "dbt": DbtCliResource(project_dir=os.fspath(bigquery_with_check_failure_jaffle_dir)),
        },
        raise_on_error=False,
    )

    asset_keys_with_metadata = set()
    for event in result.get_asset_observation_events():
        for key in event.asset_observation_data.asset_observation.metadata.keys():
            if key == BIGQUERY_METADATA_BYTES_BILLED or key == BIGQUERY_METADATA_SLOTS_MS:
                asset_keys_with_metadata.add(
                    event.asset_observation_data.asset_observation.asset_key.to_user_string()
                )

    assert asset_keys_with_metadata == JAFFLE_SHOP_ASSET_KEYS
    assert not result.success


def test_bigquery_with_execution_project(
    bigquery_with_execution_project_manifest_path,
    bigquery_with_execution_project_jaffle_dir,
    bigquery_client_with_execution_project,
):
    @dbt_assets(manifest=bigquery_with_execution_project_manifest_path)
    def jaffle_shop_dbt_assets(
        context: AssetExecutionContext,
        dbt: DbtCliResource,
    ):
        dbt_cli_invocation = dbt.cli(["build", "--profile", "unittest"], context=context)
        yield from dbt_with_bigquery_insights(context, dbt_cli_invocation)

    result = materialize(
        [jaffle_shop_dbt_assets],
        resources={
            "dbt": DbtCliResource(
                project_dir=os.fspath(bigquery_with_execution_project_jaffle_dir)
            ),
        },
        raise_on_error=False,
    )

    asset_keys_with_metadata = set()
    for event in result.get_asset_observation_events():
        for key in event.asset_observation_data.asset_observation.metadata.keys():
            if key == BIGQUERY_METADATA_BYTES_BILLED or key == BIGQUERY_METADATA_SLOTS_MS:
                asset_keys_with_metadata.add(
                    event.asset_observation_data.asset_observation.asset_key.to_user_string()
                )

    assert asset_keys_with_metadata == JAFFLE_SHOP_ASSET_KEYS
    assert not result.success

    # Assert that the query was executed in the execution project, not the default project
    assert (
        "`fake_exec_project`.`region-us`.INFORMATION_SCHEMA.JOBS"
        in bigquery_client_with_execution_project.query.call_args[0][0]
    )



from pathlib import Path
from typing import TYPE_CHECKING, Any

import dagster._check as check
from dagster._core.storage.defs_state.base import DefsStateStorage
from dagster._serdes import ConfigurableClass, ConfigurableClassData
from dagster_cloud_cli.core.artifacts import download_artifact, upload_artifact
from dagster_cloud_cli.core.headers.auth import DagsterCloudInstanceScope
from dagster_shared.serdes import deserialize_value
from dagster_shared.serdes.objects.models.defs_state_info import DefsStateInfo
from typing_extensions import Self

if TYPE_CHECKING:
    from dagster_cloud.instance import DagsterCloudAgentInstance  # noqa: F401

GET_LATEST_DEFS_STATE_INFO_QUERY = """
    query getLatestDefsStateInfo {
        latestDefsStateInfo
    }
"""

SET_LATEST_VERSION_MUTATION = """
    mutation setLatestDefsStateVersion($key: String!, $version: String!) {
        defsState {
            setLatestDefsStateVersion(key: $key, version: $version) {
                ok
            }
        }
    }
"""


class GraphQLDefsStateStorage(DefsStateStorage["DagsterCloudAgentInstance"], ConfigurableClass):
    def __init__(
        self, inst_data: ConfigurableClassData | None = None, override_graphql_client=None
    ):
        """Initialize this class directly only for test (using `override_graphql_client`).
        Use the ConfigurableClass machinery to init from instance yaml.
        """
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self._override_graphql_client = override_graphql_client

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: Any) -> Self:
        return cls(inst_data=inst_data)

    @property
    def url(self) -> str:
        return self._instance.dagster_cloud_url

    @property
    def api_token(self) -> str:
        return check.not_none(self._instance.dagster_cloud_agent_token)

    @property
    def deployment(self) -> str:
        return check.not_none(self._instance.deployment_name)

    @property
    def graphql_client(self):
        return (
            self._override_graphql_client
            if self._override_graphql_client
            else self._instance.graphql_client
        )

    def _execute_query(self, query, variables=None, idempotent_mutation=False):
        return self.graphql_client.execute(
            query, variable_values=variables, idempotent_mutation=idempotent_mutation
        )

    def _get_artifact_key(self, key: str, version: str) -> str:
        return f"__state__/{self._sanitize_key(key)}/{version}"

    def download_state_to_path(self, key: str, version: str, path: Path) -> None:
        download_artifact(
            url=self.url,
            scope=DagsterCloudInstanceScope.DEPLOYMENT,
            api_token=self.api_token,
            key=self._get_artifact_key(key, version),
            path=path,
            deployment=self.deployment,
        )

    def upload_state_from_path(self, key: str, version: str, path: Path) -> None:
        upload_artifact(
            url=self.url,
            scope=DagsterCloudInstanceScope.DEPLOYMENT,
            api_token=self.api_token,
            key=self._get_artifact_key(key, version),
            path=path,
            deployment=self.deployment,
        )
        self.set_latest_version(key, version)

    def get_latest_defs_state_info(self) -> DefsStateInfo | None:
        res = self._execute_query(GET_LATEST_DEFS_STATE_INFO_QUERY)
        result = res["data"]["latestDefsStateInfo"]
        if result is not None:
            return deserialize_value(result, DefsStateInfo)
        else:
            return None

    def set_latest_version(self, key: str, version: str) -> None:
        self._execute_query(SET_LATEST_VERSION_MUTATION, variables={"key": key, "version": version})
